# Copyright (c) 2018-2020 Patricio Cubillos.
# bibmanager is open-source software under the MIT license (see LICENSE).

import os
import datetime
import pickle
import shutil
import pathlib
import pytest

import bibmanager as bibm
import bibmanager.utils as u
import bibmanager.bib_manager as bm


def test_Bib_minimal(entries):
    # Minimal entry (key, author, title, and year):
    bib = bm.Bib(entries['jones_minimal'])
    assert bib.content == entries['jones_minimal']
    assert bib.key  == "JonesEtal2001scipy"
    assert bib.authors == [
        u.Author(last='Jones', first='Eric', von='', jr=''),
        u.Author(last='Oliphant', first='Travis', von='', jr=''),
        u.Author(last='Peterson', first='Pearu', von='', jr='')]
    assert bib.sort_author == u.Sort_author(last='jones', first='e',
                                  von='', jr='', year=2001, month=13)
    assert bib.year == 2001
    assert bib.title == "SciPy: Open source scientific tools for Python"
    assert bib.doi == None
    assert bib.bibcode == None
    assert bib.adsurl == None
    assert bib.eprint == None
    assert bib.isbn == None
    assert bib.month == 13
    assert bib.pdf is None
    assert bib.freeze is None


def test_Bib_ads_entry(entries):
    # Entry with more fields:
    bib = bm.Bib(entries['sing'])
    assert bib.content == entries['sing']
    assert bib.key  == "SingEtal2016natHotJupiterTransmission"
    assert bib.authors == [
        u.Author(last='{Sing}', first='D. K.', von='', jr=''),
        u.Author(last='{Fortney}', first='J. J.', von='', jr=''),
        u.Author(last='{Nikolov}', first='N.', von='', jr=''),
        u.Author(last='{Wakeford}', first='H. R.', von='', jr=''),
        u.Author(last='{Kataria}', first='T.', von='', jr=''),
        u.Author(last='{Evans}', first='T. M.', von='', jr=''),
        u.Author(last='{Aigrain}', first='S.', von='', jr=''),
        u.Author(last='{Ballester}', first='G. E.', von='', jr=''),
        u.Author(last='{Burrows}', first='A. S.', von='', jr=''),
        u.Author(last='{Deming}', first='D.', von='', jr=''),
        u.Author(last="{D{'e}sert}", first='J.-M.', von='', jr=''),
        u.Author(last='{Gibson}', first='N. P.', von='', jr=''),
        u.Author(last='{Henry}', first='G. W.', von='', jr=''),
        u.Author(last='{Huitson}', first='C. M.', von='', jr=''),
        u.Author(last='{Knutson}', first='H. A.', von='', jr=''),
        u.Author(last='{Lecavelier Des Etangs}', first='A.', von='', jr=''),
        u.Author(last='{Pont}', first='F.', von='', jr=''),
        u.Author(last='{Showman}', first='A. P.', von='', jr=''),
        u.Author(last='{Vidal-Madjar}', first='A.', von='', jr=''),
        u.Author(last='{Williamson}', first='M. H.', von='', jr=''),
        u.Author(last='{Wilson}', first='P. A.', von='', jr='')]
    assert bib.sort_author == u.Sort_author(last='sing', first='dk', von='',
                                  jr='', year=2016, month=1)
    assert bib.year == 2016
    assert bib.title == "A continuum from clear to cloudy hot-Jupiter exoplanets without primordial water depletion"
    assert bib.doi == "10.1038/nature16068"
    assert bib.bibcode == "2016Natur.529...59S"
    assert bib.adsurl == "http://adsabs.harvard.edu/abs/2016Natur.529...59S"
    assert bib.eprint == "1512.04341"
    assert bib.isbn == None
    assert bib.month == 1


def test_Bib_update_content(entries):
    bib1 = bm.Bib(entries['jones_minimal'])
    bib1.bibcode = 'bibcode1'
    bib1.pdf = 'pdf1'
    bib1.freeze = True
    bib2 = bm.Bib(entries['jones_minimal'])
    bib2.pdf = 'pdf2'
    bib1.update_content(bib2)
    # bibcode gets updated to None since it's bibtex info:
    assert bib1.bibcode is None
    # pdf gets updated since it's not None:
    assert bib1.pdf == 'pdf2'
    # freeze does not get updated, since it's not bibtex and is None:
    assert bib1.freeze is True


def test_Bib_year_raise(entries):
    # No year:
    with pytest.raises(ValueError, match="Bibtex entry 'JonesEtal2001scipy' is"
                                         " missing author, title, or year."):
        bib = bm.Bib(entries['jones_no_year'])


def test_Bib_title_raise(entries):
    # No title:
    with pytest.raises(ValueError, match="Bibtex entry 'JonesEtal2001scipy' is"
                                         " missing author, title, or year."):
        bib = bm.Bib(entries['jones_no_title'])


def test_Bib_author_raise(entries):
    # No author:
    with pytest.raises(ValueError, match="Bibtex entry 'JonesEtal2001scipy' is"
                                         " missing author, title, or year."):
        bib = bm.Bib(entries['jones_no_author'])


def test_Bib_braces_raise(entries):
    # Mismatched braces:
    with pytest.raises(ValueError, match="Mismatched braces in entry."):
        bib = bm.Bib(entries['jones_braces'])


def test_Bib_update_key(entries):
    bib = bm.Bib(entries['jones_minimal'])
    assert bib.key  == "JonesEtal2001scipy"
    bib.update_key("JonesOliphantPeterson2001scipy")
    assert bib.key == "JonesOliphantPeterson2001scipy"
    assert bib.content == '''\
@Misc{JonesOliphantPeterson2001scipy,
  author = {Eric Jones and Travis Oliphant and Pearu Peterson},
  title  = {{SciPy}: Open source scientific tools for {Python}},
  year   = {2001},
}'''


def test_Bib_contains(bibs):
    bib = bm.Bib('''@ARTICLE{DoeEtal2020,
                    author = {{Doe}, J. and {Perez}, J. and {Dupont}, J.},
                     title = "What Have the Astromomers ever Done for Us?",
                      year = 2020,}''')
    assert 'Doe, J'   in bib
    assert 'John Doe' in bib
    assert 'Doe'      in bib
    assert 'Doe, K.'  not in bib
    assert 'Doe, J K' not in bib
    assert '^Doe'     in bib
    assert '^J Doe'   in bib
    assert '^Perez'   not in bib


def test_Bib_published_peer_reviewed():
    # Has adsurl field, no 'arXiv' in it:
    entry = '''@ARTICLE{DoeEtal2020,
          author = {{Doe}, J. and {Perez}, J. and {Dupont}, J.},
           title = "What Have the Astromomers ever Done for Us?",
          adsurl = {http://adsabs.harvard.edu/abs/2016Natur.123...45S},
            year = 2020,}'''
    bib = bm.Bib(entry)
    assert bib.published() == 1


def test_Bib_published_arxiv():
    # Has 'arXiv' adsurl:
    entry = '''@ARTICLE{DoeEtal2020,
          author = {{Doe}, J. and {Perez}, J. and {Dupont}, J.},
           title = "What Have the Astromomers ever Done for Us?",
          adsurl = {http://adsabs.harvard.edu/abs/2016arXiv0123.0045S},
            year = 2020,}'''
    bib = bm.Bib(entry)
    assert bib.published() == 0


def test_Bib_published_non_ads():
    # Does not have adsurl:
    entry = '''@ARTICLE{DoeEtal2020,
          author = {{Doe}, J. and {Perez}, J. and {Dupont}, J.},
           title = "What Have the Astromomers ever Done for Us?",
            year = 2020,}'''
    bib = bm.Bib(entry)
    assert bib.published() == -1


def test_Bib_meta():
    e = '''@Misc{JonesEtal2001scipy,
       author = {Eric Jones},
       title  = {SciPy},
       year   = {2001},
    }'''
    bib = bm.Bib(e)
    assert bib.meta() == ''
    bib = bm.Bib(e, freeze=True, pdf='file.pdf')
    assert bib.meta() == 'freeze\npdf: file.pdf\n'


def test_display_bibs(capfd, mock_init):
    e1 = '''@Misc{JonesEtal2001scipy,
       author = {Eric Jones},
       title  = {SciPy},
       year   = {2001},
    }'''
    e2 = '''@Misc{Jones2001,
       author = {Travis Oliphant},
       title  = {tools for Python},
       year   = {2001},
    }'''
    bibs = [bm.Bib(e1), bm.Bib(e2)]
    bm.display_bibs(["DATABASE:\n", "NEW:\n"], bibs)
    captured = capfd.readouterr()
    assert captured.out == '\x1b[0m\x1b[?7h\x1b[0;38;5;248;3m\r\n::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::\r\n\x1b[0mDATABASE:\r\n\x1b[0;38;5;34;1;4m@Misc\x1b[0m{\x1b[0;38;5;142mJonesEtal2001scipy\x1b[0m,\x1b[0m\r\n       \x1b[0;38;5;33mauthor\x1b[0m \x1b[0m=\x1b[0m \x1b[0;38;5;130m{\x1b[0;38;5;130mEric Jones\x1b[0;38;5;130m}\x1b[0m,\x1b[0m\r\n       \x1b[0;38;5;33mtitle\x1b[0m  \x1b[0m=\x1b[0m \x1b[0;38;5;130m{\x1b[0;38;5;130mSciPy\x1b[0;38;5;130m}\x1b[0m,\x1b[0m\r\n       \x1b[0;38;5;33myear\x1b[0m   \x1b[0m=\x1b[0m \x1b[0;38;5;130m{\x1b[0;38;5;130m2001\x1b[0;38;5;130m}\x1b[0m,\x1b[0m\r\n    \x1b[0m}\x1b[0m\r\n\x1b[0m\r\n\x1b[0mNEW:\r\n\x1b[0;38;5;34;1;4m@Misc\x1b[0m{\x1b[0;38;5;142mJones2001\x1b[0m,\x1b[0m\r\n       \x1b[0;38;5;33mauthor\x1b[0m \x1b[0m=\x1b[0m \x1b[0;38;5;130m{\x1b[0;38;5;130mTravis Oliphant\x1b[0;38;5;130m}\x1b[0m,\x1b[0m\r\n       \x1b[0;38;5;33mtitle\x1b[0m  \x1b[0m=\x1b[0m \x1b[0;38;5;130m{\x1b[0;38;5;130mtools for Python\x1b[0;38;5;130m}\x1b[0m,\x1b[0m\r\n       \x1b[0;38;5;33myear\x1b[0m   \x1b[0m=\x1b[0m \x1b[0;38;5;130m{\x1b[0;38;5;130m2001\x1b[0;38;5;130m}\x1b[0m,\x1b[0m\r\n    \x1b[0m}\x1b[0m\r\n\x1b[0m\r\n\x1b[0m\x1b[0m'


def test_display_bibs_with_meta(capfd, mock_init):
    e1 = '''@Misc{JonesEtal2001scipy,
       author = {Eric Jones},
       title  = {SciPy},
       year   = {2001},
    }'''
    e2 = '''@Misc{Jones2001,
       author = {Travis Oliphant},
       title  = {tools for Python},
       year   = {2001},
    }'''
    bibs = [bm.Bib(e1), bm.Bib(e2, freeze=True, pdf='file.pdf')]
    bm.display_bibs(["DATABASE:\n", "NEW:\n"], bibs)
    captured = capfd.readouterr()
    assert captured.out == '\x1b[0m\x1b[?7h\x1b[0;38;5;248;3m\r\n::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::\r\n\x1b[0mDATABASE:\r\n\x1b[0;38;5;34;1;4m@Misc\x1b[0m{\x1b[0;38;5;142mJonesEtal2001scipy\x1b[0m,\x1b[0m\r\n       \x1b[0;38;5;33mauthor\x1b[0m \x1b[0m=\x1b[0m \x1b[0;38;5;130m{\x1b[0;38;5;130mEric Jones\x1b[0;38;5;130m}\x1b[0m,\x1b[0m\r\n       \x1b[0;38;5;33mtitle\x1b[0m  \x1b[0m=\x1b[0m \x1b[0;38;5;130m{\x1b[0;38;5;130mSciPy\x1b[0;38;5;130m}\x1b[0m,\x1b[0m\r\n       \x1b[0;38;5;33myear\x1b[0m   \x1b[0m=\x1b[0m \x1b[0;38;5;130m{\x1b[0;38;5;130m2001\x1b[0;38;5;130m}\x1b[0m,\x1b[0m\r\n    \x1b[0m}\x1b[0m\r\n\x1b[0m\r\n\x1b[0mNEW:\r\n\x1b[0;38;5;34;1;4m@Misc\x1b[0m{\x1b[0;38;5;142mJones2001\x1b[0m,\x1b[0m\r\n       \x1b[0;38;5;33mauthor\x1b[0m \x1b[0m=\x1b[0m \x1b[0;38;5;130m{\x1b[0;38;5;130mTravis Oliphant\x1b[0;38;5;130m}\x1b[0m,\x1b[0m\r\n       \x1b[0;38;5;33mtitle\x1b[0m  \x1b[0m=\x1b[0m \x1b[0;38;5;130m{\x1b[0;38;5;130mtools for Python\x1b[0;38;5;130m}\x1b[0m,\x1b[0m\r\n       \x1b[0;38;5;33myear\x1b[0m   \x1b[0m=\x1b[0m \x1b[0;38;5;130m{\x1b[0;38;5;130m2001\x1b[0;38;5;130m}\x1b[0m,\x1b[0m\r\n    \x1b[0m}\x1b[0m\r\n\x1b[0m\r\n\x1b[0m\x1b[0m'

    bm.display_bibs(["DATABASE:\n", "NEW:\n"], bibs, meta=True)
    captured = capfd.readouterr()
    assert captured.out == '\x1b[0m\x1b[?7h\x1b[0;38;5;248;3m\r\n::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::\r\n\x1b[0mDATABASE:\r\n\x1b[0;38;5;248;3m\x1b[0;38;5;34;1;4m@Misc\x1b[0m{\x1b[0;38;5;142mJonesEtal2001scipy\x1b[0m,\x1b[0m\r\n       \x1b[0;38;5;33mauthor\x1b[0m \x1b[0m=\x1b[0m \x1b[0;38;5;130m{\x1b[0;38;5;130mEric Jones\x1b[0;38;5;130m}\x1b[0m,\x1b[0m\r\n       \x1b[0;38;5;33mtitle\x1b[0m  \x1b[0m=\x1b[0m \x1b[0;38;5;130m{\x1b[0;38;5;130mSciPy\x1b[0;38;5;130m}\x1b[0m,\x1b[0m\r\n       \x1b[0;38;5;33myear\x1b[0m   \x1b[0m=\x1b[0m \x1b[0;38;5;130m{\x1b[0;38;5;130m2001\x1b[0;38;5;130m}\x1b[0m,\x1b[0m\r\n    \x1b[0m}\x1b[0m\r\n\x1b[0m\r\n\x1b[0mNEW:\r\n\x1b[0;38;5;248;3mfreeze\r\npdf: file.pdf\r\n\x1b[0;38;5;34;1;4m@Misc\x1b[0m{\x1b[0;38;5;142mJones2001\x1b[0m,\x1b[0m\r\n       \x1b[0;38;5;33mauthor\x1b[0m \x1b[0m=\x1b[0m \x1b[0;38;5;130m{\x1b[0;38;5;130mTravis Oliphant\x1b[0;38;5;130m}\x1b[0m,\x1b[0m\r\n       \x1b[0;38;5;33mtitle\x1b[0m  \x1b[0m=\x1b[0m \x1b[0;38;5;130m{\x1b[0;38;5;130mtools for Python\x1b[0;38;5;130m}\x1b[0m,\x1b[0m\r\n       \x1b[0;38;5;33myear\x1b[0m   \x1b[0m=\x1b[0m \x1b[0;38;5;130m{\x1b[0;38;5;130m2001\x1b[0;38;5;130m}\x1b[0m,\x1b[0m\r\n    \x1b[0m}\x1b[0m\r\n\x1b[0m\r\n\x1b[0m\x1b[0m'


def test_remove_duplicates_no_duplicates(bibs):
    # No duplicates, no removal:
    my_bibs = [bibs['beaulieu_apj'], bibs['stodden']]
    bm.remove_duplicates(my_bibs, "doi")
    assert len(my_bibs) == 2


def test_remove_duplicates_identical(bibs):
    # Identical entries:
    my_bibs = [bibs["beaulieu_apj"], bibs["beaulieu_apj"]]
    bm.remove_duplicates(my_bibs, "doi")
    assert my_bibs == [bibs["beaulieu_apj"]]


def test_remove_duplicates_diff_published(bibs):
    # Duplicate, differente published status
    my_bibs = [bibs["beaulieu_apj"], bibs["beaulieu_arxiv"]]
    bm.remove_duplicates(my_bibs, "eprint")
    assert len(my_bibs) == 1
    assert my_bibs == [bibs["beaulieu_apj"]]


@pytest.mark.parametrize('mock_input', [['2']], indirect=True)
def test_remove_duplicates_query(bibs, mock_input, mock_init):
    # Querry-solve duplicate:
    my_bibs = [bibs["beaulieu_arxiv"], bibs["beaulieu_arxiv_dup"]]
    bm.remove_duplicates(my_bibs, "eprint")
    assert len(my_bibs) == 1
    # Note that the mocked input '2' applies on the sorted entries
    # (which, in fact, has swapped the values as seen above in my_bibs)
    assert my_bibs == [bibs["beaulieu_arxiv"]]


def test_filter_field_no_conflict(bibs, mock_init):
    # No modification to my_bibs nor new lists:
    my_bibs = [bibs["beaulieu_apj"]]
    new     = [bibs["stodden"]]
    bm.filter_field(my_bibs, new, "doi", "old")
    assert bibs["beaulieu_apj"] in my_bibs
    assert bibs["stodden"]      in new


def test_filter_field_take_published(bibs):
    # Take from new, regardless of 'take' argument:
    my_bibs = [bibs["beaulieu_arxiv"]]
    new     = [bibs["beaulieu_apj"]]
    bm.filter_field(my_bibs, new, "eprint", "old")
    assert bibs["beaulieu_apj"] in my_bibs
    assert len(my_bibs) == 1
    assert new == []


def test_filter_field_take_old(bibs):
    # Take from old:
    my_bibs = [bibs["beaulieu_arxiv"]]
    new     = [bibs["beaulieu_arxiv_dup"]]
    bm.filter_field(my_bibs, new, "eprint", "old")
    assert bibs["beaulieu_arxiv"] in my_bibs
    assert len(my_bibs) == 1
    assert new == []


def test_filter_field_take_new(bibs):
    # Take from new:
    my_bibs = [bibs["beaulieu_arxiv"]]
    new     = [bibs["beaulieu_arxiv_dup"]]
    bm.filter_field(my_bibs, new, "eprint", "new")
    assert bibs["beaulieu_arxiv_dup"] in my_bibs
    assert len(my_bibs) == 1
    assert new == []


@pytest.mark.parametrize('mock_input', [['']], indirect=True)
def test_filter_field_take_ask(bibs, mock_input, mock_init):
    # Ask, keep old:
    my_bibs = [bibs["beaulieu_arxiv"]]
    new     = [bibs["beaulieu_arxiv_dup"]]
    bm.filter_field(my_bibs, new, "eprint", "ask")
    assert bibs["beaulieu_arxiv"] in my_bibs
    assert len(my_bibs) == 1
    assert new == []


@pytest.mark.parametrize('mock_input', [['n']], indirect=True)
def test_filter_field_take_ask2(bibs, mock_input, mock_init):
    # Ask, keep new:
    my_bibs = [bibs["beaulieu_arxiv"]]
    new     = [bibs["beaulieu_arxiv_dup"]]
    bm.filter_field(my_bibs, new, "eprint", "ask")
    assert bibs["beaulieu_arxiv_dup"] in my_bibs
    assert len(my_bibs) == 1
    assert new == []


def test_loadfile_bibfile(mock_init):
    bibs = bm.loadfile(u.ROOT+'examples/sample.bib')
    assert len(bibs) == 17


def test_loadfile_text(mock_init):
    with open(u.ROOT+'examples/sample.bib') as f:
       text = f.read()
    bibs = bm.loadfile(text=text)
    assert len(bibs) == 17


def test_loadfile_meta():
    with open(u.ROOT+'examples/sample.bib') as f:
       text = f.read()
    # prepend meta info before first entry:
    text = 'freeze\npdf: file.pdf\n'+ text
    bibs = bm.loadfile(text=text)

    assert bibs[0].pdf == 'file.pdf'
    assert bibs[0].freeze is True
    assert bibs[1].pdf is None
    assert bibs[1].freeze is None


def test_loadfile_pdf_with_path(tmp_path, mock_init):
    pdf_path = str(tmp_path) + '/pathed_file.pdf'
    pathlib.Path(pdf_path).touch()
    with open(u.ROOT+'examples/sample.bib') as f:
       text = f.read()
    text = f'pdf: {pdf_path}\n{text}'
    bibs = bm.loadfile(text=text)
    assert bibs[0].pdf == 'pathed_file.pdf'
    assert 'pathed_file.pdf' in os.listdir(u.BM_PDF())
    assert not os.path.isfile(pdf_path)


def test_loadfile_pdf_with_bad_path(tmp_path, mock_init):
    pdf_path = str(tmp_path) + '/pathed_file.pdf'
    # (no touch)
    with open(u.ROOT+'examples/sample.bib') as f:
       text = f.read()
    text = f'pdf: {pdf_path}\n{text}'
    bibs = bm.loadfile(text=text)
    assert bibs[0].pdf is None


def test_save(bibs, mock_init):
    my_bibs = [bibs["beaulieu_apj"]]
    bm.save(my_bibs)
    assert "bm_database.pickle" in os.listdir(u.HOME)


def test_load(bibs, mock_init):
    my_bibs = [bibs["beaulieu_apj"], bibs["stodden"]]
    bm.save(my_bibs)
    loaded_bibs = bm.load()
    assert loaded_bibs == my_bibs


def test_load_filed(tmp_path, bibs, mock_init):
    my_bibs = [bibs["beaulieu_apj"], bibs["stodden"]]
    bm.save(my_bibs)
    db = f'{tmp_path}/bm_database.pickle'
    shutil.copy(u.BM_DATABASE(), db)
    loaded_bibs = bm.load(db)
    assert loaded_bibs == my_bibs


def test_find_key(mock_init_sample):
    key = 'AASteamHendrickson2018aastex62'
    bib = bm.find(key=key)
    assert bib is not None
    assert bib.key == key 


def test_find_bibcode(mock_init_sample):
    bibcode = '2013A&A...558A..33A'
    bib = bm.find(bibcode=bibcode)
    assert bib is not None
    assert bib.bibcode == bibcode


def test_find_key_bibcode(mock_init_sample):
    key = 'AASteamHendrickson2018aastex62'
    bibcode = '2013A&A...558A..33A'
    bib = bm.find(key=key, bibcode=bibcode)
    assert bib is not None
    assert bib.key == key 
    assert bib.bibcode != bibcode


def test_find_key_not_found(mock_init_sample):
    bib = bm.find(key='non_existing_key')
    assert bib is None


def test_find_bibcode_not_found(mock_init_sample):
    bib = bm.find(bibcode='non_existing_bibcode')
    assert bib is None


def test_find_bibs(bibs):
    key = 'StoddenEtal2009ciseRRlegal'
    my_bibs = [bibs["beaulieu_apj"], bibs["stodden"]]
    bib = bm.find(key=key, bibs=my_bibs)
    assert bib is not None
    assert bib.key == key


def test_find_no_arguments(mock_init_sample):
    with pytest.raises(ValueError,
            match="Either key or bibcode arguments must be specified."):
        bib = bm.find()


def test_get_version_older(mock_init):
    # Mock pickle DB file without version:
    with open(u.BM_DATABASE(), 'wb') as handle:
        pickle.dump([], handle, protocol=pickle.HIGHEST_PROTOCOL)
    assert bm.get_version() == '0.0.0'


def test_get_version_no_pickle(mock_init):
    # Make sure there's no database:
    with u.ignored(OSError):
        os.remove(u.BM_DATABASE())
    assert bm.get_version() == bibm.__version__


def test_get_version_existing(mock_init):
    expected_version = '1.0.0'
    # Mock pickle DB file with version:
    with open(u.BM_DATABASE(), 'wb') as handle:
        pickle.dump([], handle, protocol=pickle.HIGHEST_PROTOCOL)
        pickle.dump(expected_version, handle, protocol=pickle.HIGHEST_PROTOCOL)
    assert bm.get_version() == expected_version


def test_get_version_filed(tmp_path, mock_init):
    expected_version = '1.0.0'
    db = f'{tmp_path}/bm_database.pickle'
    with open(db, 'wb') as handle:
        pickle.dump([], handle, protocol=pickle.HIGHEST_PROTOCOL)
        pickle.dump(expected_version, handle, protocol=pickle.HIGHEST_PROTOCOL)
    assert bm.get_version(db) == expected_version


def test_export_home(bibs, mock_init):
    my_bibs = [bibs["stodden"], bibs["beaulieu_apj"]]
    bm.export(my_bibs, u.BM_BIBFILE())
    assert "bm_bibliography.bib" in os.listdir(u.HOME)
    with open(u.BM_BIBFILE(), "r") as f:
        lines = f.readlines()
    assert lines[0] == "This file was created by bibmanager\n"
    loaded_bibs = bm.loadfile(u.BM_BIBFILE())
    assert loaded_bibs == sorted(my_bibs)


def test_export_no_overwrite(bibs, mock_init):
    with open(u.BM_BIBFILE(), "w") as f:
        f.write("placeholder file.")
    my_bibs = [bibs["beaulieu_apj"], bibs["stodden"]]
    bm.export(my_bibs, u.BM_BIBFILE())
    assert "bm_bibliography.bib" in os.listdir(u.HOME)
    assert f"orig_{datetime.date.today()}_bm_bibliography.bib" \
           in os.listdir(u.HOME)


def test_export_meta(mock_init_sample):
    bibs = bm.load()
    bm_export = u.HOME+'tmp_bm.bib'
    bm.export(bibs, bm_export, meta=True)
    with open(bm_export, "r") as f:
        lines = f.readlines()
    assert "pdf: Slipher1913.pdf\n" in lines


def test_export_no_meta(mock_init_sample):
    bibs = bm.load()
    bm_export = u.HOME+'tmp_bm.bib'
    bm.export(bibs, bm_export, meta=False)
    with open(bm_export, "r") as f:
        lines = f.readlines()
    assert "pdf: Slipher1913.pdf\n" not in lines


def test_merge_bibfile(capfd, mock_init):
    bm.merge(u.HOME + "examples/sample.bib")
    captured = capfd.readouterr()
    assert captured.out == "\nMerged 17 new entries.\n"


def test_merge_bibs(capfd, mock_init):
    new = bm.loadfile(u.HOME + "examples/sample.bib")
    bm.merge(new=new)
    captured = capfd.readouterr()
    assert captured.out == "\nMerged 17 new entries.\n"


def test_merge_no_new(capfd, bibs, mock_init_sample):
    bm.merge(new=[bibs['hunter']])
    captured = capfd.readouterr()
    assert captured.out == "\nMerged 0 new entries.\n"


def test_merge_base(bibs):
    merged = bm.merge(new=[bibs['hunter']], base=[bibs['stodden']])
    assert len(merged) == 2
    assert merged[0] == bibs['hunter']
    assert merged[1] == bibs['stodden']


@pytest.mark.parametrize('mock_input', [['n']], indirect=True)
def test_merge_duplicate_key_ingnore(bibs, mock_init_sample, mock_input):
    bm.merge(new=[bibs['oliphant_dup']])
    loaded_bibs = bm.load()
    assert len(loaded_bibs) == 17
    assert bibs['oliphant_dup'] in loaded_bibs


@pytest.mark.parametrize('mock_input', [['Oliphant2016numpyb']], indirect=True)
def test_merge_duplicate_key_rename(bibs, mock_init_sample, mock_input):
    bm.merge(new=[bibs['oliphant_dup']])
    loaded_bibs = bm.load()
    assert len(loaded_bibs) == 18
    assert 'Oliphant2016numpyb' in [e.key for e in loaded_bibs]


@pytest.mark.parametrize('mock_input', [['']], indirect=True)
def test_merge_duplicate_title_ignore(bibs, mock_init_sample, mock_input):
    bm.merge(new=[bibs['no_oliphant']])
    loaded_bibs = bm.load()
    assert len(loaded_bibs) == 17
    assert bibs['no_oliphant'] not in loaded_bibs


@pytest.mark.parametrize('mock_input', [['a']], indirect=True)
def test_merge_duplicate_title_add(bibs, mock_init_sample, mock_input):
    bm.merge(new=[bibs['no_oliphant']])
    loaded_bibs = bm.load()
    assert len(loaded_bibs) == 18
    assert bibs['no_oliphant'] in loaded_bibs


def test_init_from_scratch(mock_home):
    shutil.rmtree(u.HOME, ignore_errors=True)
    bm.init(bibfile=None)
    assert set(os.listdir(u.HOME)) == set(["config", "examples", "pdf"])

    with open(u.HOME+"config", 'r') as f:
        home = f.read()
    with open(u.ROOT+"config", 'r') as f:
        root = f.read()
    assert home == root.replace('HOME/', u.HOME)

    assert set(os.listdir(u.HOME+"examples")) \
        == set(['aastex62.cls', 'apj_hyperref.bst', 'sample.bib', 'sample.tex',
                'top-apj.tex'])


@pytest.mark.parametrize('mock_prompt', [['']], indirect=True)
def test_add_entries_dry(capfd, mock_init, mock_prompt):
    bm.add_entries('new')
    captured = capfd.readouterr()
    assert captured.out == (
        "Enter a BibTeX entry (press META+ENTER or ESCAPE ENTER when done):\n"
        "\nNo new entries to add.\n")


@pytest.mark.parametrize('mock_prompt', [['this will fail}']], indirect=True)
def test_add_entries_raise(capfd, mock_init, mock_prompt):
    with pytest.raises(ValueError,
                       match="Mismatched braces in line 0:\n'this will fail}'"):
        bm.add_entries('new')


# TBD: Can I pass a fixure to the decorator?
@pytest.mark.parametrize('mock_prompt',
 [['''@Misc{JonesEtal2001scipy,
   author = {Eric Jones and Travis Oliphant and Pearu Peterson},
   title  = {{SciPy}: Open source scientific tools for {Python}},
   year   = {2001},
   }''']], indirect=True)
def test_add_entries(capfd, mock_init, mock_prompt, entries):
    bm.add_entries('new')
    captured = capfd.readouterr()
    assert captured.out == (
        "Enter a BibTeX entry (press META+ENTER or ESCAPE ENTER when done):\n"
        "\n\nMerged 1 new entries.\n")


@pytest.mark.skip(reason='No clue how to test this')
def test_edit():
    pass


def test_search_author(mock_init_sample):
    matches = bm.search(authors="oliphant")
    assert len(matches) == 2
    keys = [m.key for m in matches]
    assert 'JonesEtal2001scipy' in keys
    assert 'Oliphant2006numpy'  in keys


def test_search_author_last_initials(mock_init_sample):
    matches = bm.search(authors="oliphant, t")
    assert len(matches) == 2
    keys = [m.key for m in matches]
    assert 'JonesEtal2001scipy' in keys
    assert 'Oliphant2006numpy'  in keys


def test_search_author_first(mock_init_sample):
    matches = bm.search(authors="^oliphant, t")
    assert len(matches) == 1
    keys = [m.key for m in matches]
    assert 'Oliphant2006numpy'  in keys


def test_search_author_multiple(mock_init_sample):
    # Multiple-author querries act with AND logic:
    matches = bm.search(authors=["oliphant, t", "jones, e"])
    assert len(matches) == 1
    keys = [m.key for m in matches]
    assert 'JonesEtal2001scipy' in keys


def test_search_author_year_title(mock_init_sample):
    # Combined-fields querries act with AND logic:
    matches = bm.search(authors="oliphant, t", year=2006, title="numpy")
    assert len(matches) == 1
    keys = [m.key for m in matches]
    assert 'Oliphant2006numpy'  in keys


def test_search_title_multiple(mock_init_sample):
    # Multiple-title querries act with AND logic:
    matches = bm.search(title=['HD 209458b', 'atmospheric circulation'])
    assert len(matches) == 1
    keys = [m.key for m in matches]
    assert 'ShowmanEtal2009apjRadGCM' in keys


def test_search_year_specific(mock_init_sample):
    matches = bm.search(authors="cubillos, p", year=2016)
    assert len(matches) == 1
    keys = [m.key for m in matches]
    assert 'Cubillos2016phdThesis' in keys


def test_search_year_range(mock_init_sample):
    matches = bm.search(authors="cubillos, p", year=[2013,2016])
    assert len(matches) == 2
    keys = [m.key for m in matches]
    assert 'CubillosEtal2013apjWASP8b' in keys
    assert 'Cubillos2016phdThesis' in keys


def test_search_bibcode(mock_init_sample):
    matches = bm.search(bibcode="2013A&A...558A..33A")
    assert len(matches) == 1
    keys = [m.key for m in matches]
    assert 'Astropycollab2013aaAstropy' in keys


def test_search_bibcode_utf8(mock_init_sample):
    # UTF8 decoding is done at run time, so this works as well:
    matches = bm.search(bibcode="2013A%26A...558A..33A")
    assert len(matches) == 1
    keys = [m.key for m in matches]
    assert 'Astropycollab2013aaAstropy' in keys


def test_search_bibcode_multiple(mock_init_sample):
    # Multiple-bibcode querries act with OR logic:
    matches = bm.search(bibcode=["2013A%26A...558A..33A","1957RvMP...29..547B"])
    assert len(matches) == 2
    keys = [m.key for m in matches]
    assert 'Astropycollab2013aaAstropy' in keys
    assert 'BurbidgeEtal1957rvmpStellarElementSynthesis' in keys


def test_search_key(mock_init_sample):
    matches = bm.search(key="BurbidgeEtal1957rvmpStellarElementSynthesis")
    assert len(matches) == 1
    keys = [m.key for m in matches]
    assert 'BurbidgeEtal1957rvmpStellarElementSynthesis' in keys


def test_search_key_multiple(mock_init_sample):
    # Multiple-key querries act with OR logic:
    matches = bm.search(key=["Astropycollab2013aaAstropy",
                             "BurbidgeEtal1957rvmpStellarElementSynthesis"])
    assert len(matches) == 2
    keys = [m.key for m in matches]
    assert 'Astropycollab2013aaAstropy' in keys
    assert 'BurbidgeEtal1957rvmpStellarElementSynthesis' in keys


@pytest.mark.parametrize('mock_prompt_session',
     [['key: BurbidgeEtal1957rvmpStellarElementSynthesis']], indirect=True)
def test_prompt_search_kw1(capsys, mock_init_sample, mock_prompt_session):
    keywords = ['key', 'bibcode']
    field = 'bibcode'
    prompt_text = ("Test search  (Press 'tab' for autocomplete):\n")
    prompt_input = bm.prompt_search(keywords, field, prompt_text)
    assert prompt_input[0] == \
        ['BurbidgeEtal1957rvmpStellarElementSynthesis', None]
    assert prompt_input[1] == [None]
    captured = capsys.readouterr()
    assert captured.out == prompt_text + '\n'


@pytest.mark.parametrize('mock_prompt_session',
     [['bibcode: 1957RvMP...29..547B']], indirect=True)
def test_prompt_search_kw2(mock_init_sample, mock_prompt_session):
    keywords = ['key', 'bibcode']
    field = 'bibcode'
    prompt_text = ("Test search  (Press 'tab' for autocomplete):\n")
    prompt_input = bm.prompt_search(keywords, field, prompt_text)
    assert prompt_input[0] == [None, '1957RvMP...29..547B']
    assert prompt_input[1] == [None]


@pytest.mark.parametrize('mock_prompt_session',
     [['bibcode: 1957RvMP...29..547B extra']], indirect=True)
def test_prompt_search_extra(mock_init_sample, mock_prompt_session):
    keywords = ['key', 'bibcode']
    field = 'bibcode'
    prompt_text = ("Test search  (Press 'tab' for autocomplete):\n")
    prompt_input = bm.prompt_search(keywords, field, prompt_text)
    assert prompt_input[0] == \
        [None, '1957RvMP...29..547B']
    assert prompt_input[1] == ['extra']


@pytest.mark.parametrize('mock_prompt_session',
     [['']], indirect=True)
def test_prompt_search_empty_prompt(mock_init_sample, mock_prompt_session):
    keywords = ['key', 'bibcode']
    field = 'bibcode'
    prompt_text = ("Test search  (Press 'tab' for autocomplete):\n")
    with pytest.raises(ValueError, match='Invalid syntax.'):
        prompt_input = bm.prompt_search(keywords, field, prompt_text)


@pytest.mark.parametrize('mock_prompt_session',
     [['bibcode:']], indirect=True)
def test_prompt_search_empty_value(mock_init_sample, mock_prompt_session):
    keywords = ['key', 'bibcode']
    field = 'bibcode'
    prompt_text = ("Test search  (Press 'tab' for autocomplete):\n")
    with pytest.raises(ValueError, match='Invalid syntax.'):
        prompt_input = bm.prompt_search(keywords, field, prompt_text)


@pytest.mark.parametrize('mock_prompt_session',
     [['bibcode: ']], indirect=True)
def test_prompt_search_blank_value(mock_init_sample, mock_prompt_session):
    keywords = ['key', 'bibcode']
    field = 'bibcode'
    prompt_text = ("Test search  (Press 'tab' for autocomplete):\n")
    with pytest.raises(ValueError, match='Invalid syntax.'):
        prompt_input = bm.prompt_search(keywords, field, prompt_text)


@pytest.mark.parametrize('mock_prompt_session',
     [['bibcode: VAL1  key: VAL2']], indirect=True)
def test_prompt_search_double_def(mock_init_sample, mock_prompt_session):
    keywords = ['key', 'bibcode']
    field = 'bibcode'
    prompt_text = ("Test search  (Press 'tab' for autocomplete):\n")
    with pytest.raises(ValueError, match='Invalid syntax.'):
        prompt_input = bm.prompt_search(keywords, field, prompt_text)


